import os
import os.path
from pyrustic.threadom import Threadom
from jupitest.host.main_host import MainHost
from jupitest.host.result import Result
from jupitest.host.reloader import Reloader
from jupitest.view.log_window import LogWindow
from jupitest.view.toolbar import Toolbar
from jupitest.view.tree import Tree
from jupitest.view.main_view import MainView
from pyrustic.manager.misc import funcs


class MainHostBuilder:
    def build(self, root_path):
        return MainHost(root_path, Reloader(), ResultBuilder())


class ResultBuilder:
    def build(self, test_id, queue):
        return Result(test_id, queue)


class LogWindowBuilder:
    def build(self, master, message):
        return LogWindow(master, message).build_wait()


class ToolbarBuilder:
    def build(self, node_id, parent, callback):
        toolbar = Toolbar(node_id, parent, callback, LogWindowBuilder())
        toolbar.build()
        return toolbar


class TreeBuilder:
    def build(self, master, callback):
        return Tree(master, callback)


class MainViewBuilder:
    def build(self, app):
        jasonix = funcs.get_manager_jasonix()
        target = jasonix.data["target"]
        tests_path = os.path.join(target, "tests")
        return MainView(app,
                        Threadom(app.root),
                        MainHostBuilder().build(target),
                        ToolbarBuilder(),
                        tests_path)



import sys


class Reloader:
    def __init__(self):
        self._state = None

    def save_state(self):
        self._state = sys.modules.copy()

    def restore_state(self):
        for x in sys.modules.copy().keys():
            if not x in self._state:
                del sys.modules[x]



from pyrustic.viewable import Viewable
import math
import traceback
import tkinter as tk


class Toolbar(Viewable):
    def __init__(self, node_id, parent, callback, log_window_builder):
        self._node_id = node_id
        self._parent = parent
        self._callback = callback
        self._log_window_builder = log_window_builder
        # States variables
        self._toolbar_visible = True
        self._is_running = False
        self._is_waiting = False
        self._time_elapsed = 0
        self._count_tests = 0
        self._count_tests_started = 0
        self._count_errors = 0
        self._count_sub_test_errors = 0
        self._count_sub_test_failures = 0
        self._count_failures = 0
        self._count_skips = 0
        self._count_unexpected_successes = 0
        self._count_expected_failures = 0
        # StringVar
        self._state_stringvar = tk.StringVar()
        self._time_elapsed_stringvar = tk.StringVar()
        self._count_tests_stringvar = tk.StringVar(value="0")
        self._percent_stringvar = tk.StringVar(value="0%")
        self._count_tests_started_stringvar = tk.StringVar()
        self._error_stringvar = tk.StringVar()
        self._sub_test_error_stringvar = tk.StringVar()
        self._sub_test_failure_stringvar = tk.StringVar()
        self._failure_stringvar = tk.StringVar()
        self._skip_stringvar = tk.StringVar()
        self._expected_failure_stringvar = tk.StringVar()
        self._unexpected_success_stringvar = tk.StringVar()
        self._live_info_stringvar = tk.StringVar()
        # IntVar
        self._failfast_intvar = tk.IntVar()
        self._verbose_intvar = tk.IntVar(value=1)
        # body
        self._body = None
        # frame
        self._frame_top = self._frame_bottom = None
        self._frame_top_left = self._frame_top_right = self._frame_bottom_1 = None
        self._frame_bottom_2 = self._frame_bottom_3 = None
        # Buttons
        self._btn_run = self._btn_rerun = self._btn_stop = self._btn_cancel = None
        self._btn_log = self._btn_clean = None
        # Checkbutton
        self._checkbutton_failfast = self._checkbutton_verbose = None
        # Labels
        self._label_state = self._label_testing_passed = self._label_testing_failed = None
        self._label_time_elapsed = self._label_count_tests = None
        self._label_count_tests_completed = self._label_percent =  None
        self._label_error = self._label_sub_test_error = self._label_failure = None
        self._label_sub_test_failure = self._label_skip = None
        self._label_expected_failure = self._label_unexpected_success = None
        self._live_info_label = None
        # raw log
        self._raw_log = []
        # log
        self._log = None

    # ========================================
    #             PROPERTIES
    # ========================================
    @property
    def is_visible(self):
        return self._toolbar_visible

    @property
    def count_tests(self):
        return self._count_tests

    @count_tests.setter
    def count_tests(self, val):
        self._count_tests = val
        val = "{} test{}".format(val, "s" if val > 1 else "")
        self._count_tests_stringvar.set(val)

    # ========================================
    #             PUBLIC METHODS
    # ========================================
    def show_or_hide(self):
        if self._toolbar_visible:
            self._body.grid_remove()
            self._toolbar_visible = False
        else:
            self._body.grid()
            self._toolbar_visible = True

    def set_running_state(self):
        self._is_running = True
        self._state_stringvar.set("Running...")
        if self._is_waiting:
            self._is_waiting = False
        # remove widgets
        self._reset_toolbar()
        # set btn stop
        self._btn_stop.grid()
        # set label state
        self._label_state.grid()
        # set label count_tests_completed
        self._label_count_tests_completed.grid()
        # set label Percent
        self._label_percent.grid()
        # set live info
        if self._verbose_intvar.get() == 1:
            self._frame_bottom.grid()
            self._frame_bottom_1.grid()
            self._live_info_label.grid()

    def set_waiting_state(self):
        self._is_waiting = True
        self._state_stringvar.set("Waiting...")
        # remove widgets
        self._reset_toolbar()
        # set btn cancel
        self._btn_cancel.grid()
        # set label Current state
        self._label_state.grid()
        # set label Count_tests
        self._label_count_tests.grid()

    def set_stop_state(self, was_successful, log):
        self._raw_log = log
        self._is_running = False
        # set label testing failed or label testing passed
        if was_successful:
            self._label_testing_passed.grid()
        else:
            self._label_testing_failed.grid()
        # set time elapsed
        self._label_time_elapsed.grid()
        # remove state label
        self._label_state.grid_remove()
        # remove frame_bottom_1
        self._frame_bottom_1.grid_remove()
        # conditionally remove frame_bottom_2 and frame_bottom_3
        remove_frame_bottom_2 = True
        remove_frame_bottom_3 = True
        for x in (self._count_errors, self._count_sub_test_errors,
                  self._count_failures, self._count_sub_test_failures):
            if x != 0:
                remove_frame_bottom_2 = False
        for x in (self._count_unexpected_successes, self._count_expected_failures,
                  self._count_skips):
            if x != 0:
                remove_frame_bottom_3 = False
        if remove_frame_bottom_2:
            self._frame_bottom_2.grid_remove()
        if remove_frame_bottom_3:
            self._frame_bottom_3.grid_remove()
        if remove_frame_bottom_2 and remove_frame_bottom_3:
            self._frame_bottom.grid_remove()
        # remove stop button
        self._btn_stop.grid_remove()
        # set rerun button
        self._btn_rerun.grid()
        # set log button
        self._btn_log.grid()
        # set clean button
        self._btn_clean.grid()


    def notify_time_elapsed(self, time_elapsed):
        self._time_elapsed = math.floor(time_elapsed * 10 ** 3) / 10 ** 3
        self._time_elapsed_stringvar.set(str("{}s".format(self._time_elapsed)))

    def notify_start_test(self, test):
        self._count_tests_started += 1
        count_tests_started = "{}/{}".format(self._count_tests_started, self._count_tests)
        self._live_info_stringvar.set(">>> " + str(test))
        self._count_tests_started_stringvar.set(count_tests_started)
        percent = int((self._count_tests_started / self._count_tests) * 100)
        percent = "{}%".format(percent)
        self._percent_stringvar.set(percent)

    def notify_stop_test(self, test):
        pass

    def notify_add_error(self, test, err):
        self._count_errors += 1
        if self._verbose_intvar.get() != 1:
            return
        count_errors = "{} error{}".format(self._count_errors,
                                           "s" if self._count_errors > 1 else "")
        self._error_stringvar.set(count_errors)
        self._frame_bottom.grid()
        self._frame_bottom_2.grid()
        self._label_error.grid()

    def notify_add_failure(self, test, err):
        self._count_failures += 1
        if self._verbose_intvar.get() != 1:
            return
        count_failures = "{} failure{}".format(self._count_failures,
                                           "s" if self._count_failures > 1 else "")
        self._failure_stringvar.set(count_failures)
        self._frame_bottom.grid()
        self._frame_bottom_2.grid()
        self._label_failure.grid()

    def notify_add_success(self, test):
        pass

    def notify_add_skip(self, test, reason):
        self._count_skips += 1
        if self._verbose_intvar.get() != 1:
            return
        self._skip_stringvar.set("{} skipped".format(self._count_skips))
        self._frame_bottom.grid()
        self._frame_bottom_3.grid()
        self._label_skip.grid()

    def notify_add_expected_failure(self, test, err):
        self._count_expected_failures += 1
        if self._verbose_intvar.get() != 1:
            return
        count_expected_failures = ("{} expected-failure{}".
                                   format(self._count_expected_failures,
                            "s" if self._count_expected_failures > 1 else ""))
        self._expected_failure_stringvar.set(count_expected_failures)
        self._frame_bottom.grid()
        self._frame_bottom_3.grid()
        self._label_expected_failure.grid()

    def notify_add_unexpected_success(self, test):
        self._count_unexpected_successes += 1
        if self._verbose_intvar.get() != 1:
            return
        count_unexpected_successes = ("{} unexpected-success{}".
                                      format(self._count_unexpected_successes,
                            "es" if self._count_unexpected_successes > 1 else ""))
        self._unexpected_success_stringvar.set(count_unexpected_successes)
        self._frame_bottom.grid()
        self._frame_bottom_3.grid()
        self._label_unexpected_success.grid()

    def notify_add_sub_test(self, test, subtest, outcome):
        self._live_info_stringvar.set("Currently on: " + str(subtest))
        if not outcome:
            return
        if issubclass(outcome[0], test.failureException):
            self._count_sub_test_failures += 1
            if self._verbose_intvar.get() != 1:
                return
            count_sub_test_failures = ("{} sub-test-failure{}"
                                       .format(self._count_sub_test_failures,
                                    "s" if self._count_sub_test_failures > 1 else ""))
            self._sub_test_failure_stringvar.set(count_sub_test_failures)
            self._frame_bottom_2.grid()
            self._label_sub_test_failure.grid()
        else:
            self._count_sub_test_errors += 1
            if self._verbose_intvar.get() != 1:
                return
            count_sub_test_errors = ("{} sub-test-error{}"
                                       .format(self._count_sub_test_errors,
                                               "s" if self._count_sub_test_errors > 1 else ""))
            self._sub_test_error_stringvar.set(count_sub_test_errors)
            self._frame_bottom_2.grid()
            self._label_sub_test_error.grid()


    # ========================================
    #       VIEWABLE METHODS IMPLEMENTATION
    # ========================================

    def _on_build(self):
        self._body = tk.Frame(self._parent, name="runtestToolbar")
        self._body.grid(pady=5, sticky="w")
        # ============== frames ================
        self._frame_top = tk.Frame(self._body)
        self._frame_top_left = tk.Frame(self._frame_top)
        self._frame_top_right = tk.Frame(self._frame_top)
        self._frame_bottom = tk.Frame(self._body)
        self._frame_bottom_1 = tk.Frame(self._frame_bottom)
        self._frame_bottom_2 = tk.Frame(self._frame_bottom)
        self._frame_bottom_3 = tk.Frame(self._frame_bottom)
        # frame config
        # grid frames
        self._frame_top.grid(row=0, column=0, sticky="w")
        self._frame_top_left.grid(row=0, column=0, sticky="w")
        self._frame_top_right.grid(row=0, column=1, sticky="w")
        self._frame_bottom.grid(row=1, column=0, sticky="w")
        self._frame_bottom_1.grid(row=1, column=0, sticky="w",
                                  pady=(5, 0), padx=(0, 10))
        self._frame_bottom_2.grid(row=2, column=0, sticky="w",
                                  pady=(5, 0))
        self._frame_bottom_3.grid(row=3, column=0, sticky="w",
                                  pady=(5, 0))
        # =========== buttons =============
        # btn run
        self._btn_run = tk.Button(self._frame_top_left,
                                  name="buttonRun",
                                  text="Run",
                                  command=self._on_run_clicked)
        # btn rerun
        self._btn_rerun = tk.Button(self._frame_top_left,
                                    name="buttonRerun",
                                    text="Rerun",
                                    command=self._on_rerun_clicked)
        # btn stop
        self._btn_stop = tk.Button(self._frame_top_left,
                                   name="buttonStop",
                                   text="Stop",
                                   command=self._on_stop_clicked)
        # btn cancel
        self._btn_cancel = tk.Button(self._frame_top_left,
                                     name="buttonCancel",
                                     text="Cancel",
                                     command=self._on_cancel_clicked)
        # btn log
        self._btn_log = tk.Button(self._frame_top_left,
                                  name="buttonLog",
                                  text="Log",
                                  command=self._on_log_clicked)
        # btn clean
        self._btn_clean = tk.Button(self._frame_top_left,
                                    name="buttonClean",
                                    text="x",
                                    command=self._on_clean_clicked)
        # grid buttons
        for i, x in enumerate((self._btn_run, self._btn_rerun,
                            self._btn_stop, self._btn_cancel,
                            self._btn_log, self._btn_clean)):
            x.grid(row=0, column=i, padx=(0, 10))
        # =============== right-side of top frame ============
        # label state
        self._label_state = tk.Label(self._frame_top_right,
                                     textvariable=self._state_stringvar)
        # label testing passed
        self._label_testing_passed = tk.Label(self._frame_top_right,
                                              name="labelTestingPassed",
                                              text="Testing passed !")
        # label testing failed
        self._label_testing_failed = tk.Label(self._frame_top_right,
                                              name="labelTestingFailed",
                                              text="Testing failed !")
        # label duration
        self._label_time_elapsed = tk.Label(self._frame_top_right, textvariable=self._time_elapsed_stringvar)
        # label count_tests
        self._label_count_tests = tk.Label(self._frame_top_right,
                                           textvariable=self._count_tests_stringvar)
        # label count_tests_completed
        self._label_count_tests_completed = tk.Label(self._frame_top_right,
                                                     textvariable=self._count_tests_started_stringvar)
        # label percent
        self._label_percent = tk.Label(self._frame_top_right, textvariable=self._percent_stringvar)
        # checkbutton failfast
        self._checkbutton_failfast = tk.Checkbutton(self._frame_top_right,
                                                    text="Fail-Fast",
                                                    variable=self._failfast_intvar,
                                                    onvalue=1, offvalue=0)
        # checkbutton verbose
        self._checkbutton_verbose = tk.Checkbutton(self._frame_top_right,
                                                   text="Verbose",
                                                   variable=self._verbose_intvar,
                                                   onvalue=1, offvalue=0)
        # grid
        for i, x in enumerate((self._label_state, self._label_testing_passed,
                               self._label_testing_failed, self._label_time_elapsed,
                               self._label_count_tests, self._label_count_tests_completed,
                               self._label_percent, self._checkbutton_failfast,
                               self._checkbutton_verbose)):
            x.grid(row=0, column=i, padx=(0, 10))
        # =============== second stage frame elements ==============
        # label live info
        self._live_info_label = tk.Label(self._frame_bottom_1, textvariable=self._live_info_stringvar)
        # grid
        self._live_info_label.grid(row=0, column=0)
        # =============== third stage frame elements ==============
        # label error
        self._label_error = tk.Label(self._frame_bottom_2, textvariable=self._error_stringvar)
        # label sub_test error
        self._label_sub_test_error = tk.Label(self._frame_bottom_2,
                                              textvariable=self._sub_test_error_stringvar)
        # label failure
        self._label_failure = tk.Label(self._frame_bottom_2,
                                       textvariable=self._failure_stringvar)
        # label sub_test failure
        self._label_sub_test_failure = tk.Label(self._frame_bottom_2,
                                                textvariable=self._sub_test_failure_stringvar)
        # grid
        for i, x in enumerate((self._label_error, self._label_sub_test_error,
                               self._label_failure, self._label_sub_test_failure)):
            x.grid(row=0, column=i, padx=(0, 20))
        # =============== fourth stage frame elements ==============
        # label unexpected success
        self._label_unexpected_success = tk.Label(
            self._frame_bottom_3, textvariable=self._unexpected_success_stringvar)
        # label expected failure
        self._label_expected_failure = tk.Label(
            self._frame_bottom_3, textvariable=self._expected_failure_stringvar)
        # label skip
        self._label_skip = tk.Label(self._frame_bottom_3, textvariable=self._skip_stringvar)
        # grid
        for i, x in enumerate((self._label_unexpected_success, self._label_expected_failure,
                               self._label_skip)):
            x.grid(row=0, column=i, padx=(0, 20))
        # ====================================================
        # reset
        self._reset_toolbar()
        self._install_default_widgets()
        return self._body

    def _on_display(self):
        pass

    def _on_destroy(self):
        pass

    # ========================================
    #             PRIVATE METHODS
    # ========================================
    def _on_run_clicked(self):
        self._callback.on_run_clicked(self._node_id,
                                      self._failfast_intvar.get())

    def _on_rerun_clicked(self):
        self._on_clean_clicked()
        self._callback.on_run_clicked(self._node_id, self._failfast_intvar.get())

    def _on_stop_clicked(self):
        self._callback.on_stop_clicked(self._node_id)

    def _on_cancel_clicked(self):
        self._is_waiting = False
        self._reset_toolbar()
        self._install_default_widgets()
        self._callback.on_cancel_clicked(self._node_id)

    def _on_clean_clicked(self):
        self._reset_data()
        self._reset_toolbar()
        self._install_default_widgets()

    def _reset_toolbar(self):
        for x in (*self._frame_top_left.winfo_children(),
                  *self._frame_top_right.winfo_children(),
                  *self._frame_bottom_1.winfo_children(),
                  *self._frame_bottom_2.winfo_children(),
                  *self._frame_bottom_3.winfo_children(),
                  self._frame_bottom_1,
                  self._frame_bottom_2,
                  self._frame_bottom_3,
                  self._frame_bottom):
            x.grid_remove()

    def _reset_data(self):
        self._raw_log = []
        self._log = None
        self._percent_stringvar.set("0%")
        self._state_stringvar.set("")
        self._live_info_stringvar.set("")
        self._time_elapsed = 0
        self._count_tests_started = 0
        self._count_errors = 0
        self._count_sub_test_errors = 0
        self._count_sub_test_failures = 0
        self._count_failures = 0
        self._count_skips = 0
        self._count_unexpected_successes = 0
        self._count_expected_failures = 0

    def _install_default_widgets(self):
        # install btn run, label count_tests, checkbuttons failfast and verbose
        self._btn_run.grid()
        self._label_count_tests.grid()
        self._checkbutton_failfast.grid()
        self._checkbutton_verbose.grid()

    def _on_log_clicked(self):
        log_window = self._log_window_builder.build(self._body, self._get_log())

    def _get_log(self):
        if self._log is not None:
            return self._log
        text = ""
        line_separator = "\n\n{}\n\n".format("-" * 75)
        format_error = lambda err: "".join(traceback.format_exception(*err))
        for x in self._raw_log:
            event = x["event"]
            if event == "start_test_run":
                text += (" "*25) + "===== START TEST =====\n"
                text += line_separator
            elif event == "stop_test_run":
                text += (" "*25) + "===== STOP TEST =====\n"
            elif event == "add_error":
                text += "*** ERROR\n"
                text += "{}\n\n".format(x["test"])
                text += format_error(x["err"])
                text += line_separator
            elif event == "add_failure":
                text += "*** FAILURE\n"
                text += "{}\n\n".format(x["test"])
                text += format_error(x["err"])
                text += line_separator
            elif event == "add_skip":
                text += "*** SKIP\n"
                text += "{}\n\n".format(x["test"])
                text += "Reason: " + x["reason"]
                text += line_separator
            elif event == "add_expected_failure":
                text += "*** EXPECTED FAILURE\n"
                text += "{}\n\n".format(x["test"])
                text += format_error(x["err"])
                text += line_separator
            elif event == "add_unexpected_success":
                text += "*** UNEXPECTED SUCCESS\n"
                text += "{}\n\n".format(x["test"])
                text += line_separator
            elif event == "add_sub_test":
                outcome = x["outcome"]
                test = x["test"]
                if outcome is not None:
                    if issubclass(outcome[0], test.failureException):
                        text += "*** SUB-TEST FAILURE\n"
                    else:
                        text += "*** SUB-TEST ERROR\n"
                    text += "{}\n\n".format(x["sub_test"])
                    text += format_error(outcome)
                    text += line_separator
        self._log = text
        return self._log



from pyrustic.app import App
from jupitest.misc.builder import MainViewBuilder
from jupitest.misc import my_theme


def main():
    app = App(__package__)
    app.root.title("Jupitest - Pyrustic Test Runner")
    app.theme = my_theme.get_theme()
    app.view = MainViewBuilder().build(app)
    app.center()
    app.start()

if __name__ == "__main__":
    main()



from pyrustic.viewable import Viewable
import tkinter as tk


class LogWindow(Viewable):
    def __init__(self, master, message):
        self._master = master
        self._message = message
        self._text = None

    def _on_build(self):
        self._body = tk.Toplevel()
        self._body.title("Log - Pyrustic Test Runner")
        scrollbar = tk.Scrollbar(self._body)
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self._text = tk.Text(self._body, name="logWindow",
                             width=75, height=25)
        self._text.pack(side=tk.RIGHT, expand=1, fill=tk.BOTH)
        scrollbar.config(command=self._text.yview)
        self._text.config(yscrollcommand=scrollbar.set)
        return self._body

    def _on_display(self):
        self._text.insert("0.0", self._message)
        self._text.config(state="disabled")

    def _on_destroy(self):
        pass



import os
import os.path
import unittest
import time
from pyrustic.manager import constant


class MainHost:

    def __init__(self, root_path, reloader, result_builder):
        self._root_path = root_path
        self._reloader = reloader
        self._result_builder = result_builder
        self._is_running = False
        self._running_test = dict()
        self._setup()
        self._reloader.save_state()


    def count_tests(self, path, class_name=None, method_name=None, test_loader=None):
        self._reloader.restore_state()
        if not os.path.exists(path):
            return
        dotted_module_name = None
        test_loader = test_loader if test_loader else unittest.TestLoader()
        if os.path.isfile(path):
            dirname, ext = os.path.splitext(path)
            if ext != ".py":
                return
            dotted_module_name = self._convert_path_to_dotted_path(dirname, start="tests")
        suite = None
        try:
            if dotted_module_name:
                dotted_module_name += "." + class_name if class_name else ""
                dotted_module_name += "." + method_name if method_name else ""
                suite = test_loader.loadTestsFromName(name=dotted_module_name)
            else:
                suite = test_loader.discover(start_dir=path, top_level_dir=self._root_path)
        except Exception:
            return
        if suite is None:
            return
        return suite.countTestCases(), suite

    def tests_in_directory(self, path):
        if not os.path.exists(path):
            return
        directories = []
        modules = []
        count_tests = 0
        for name in sorted(os.listdir(path)):
            if name in ("__init__.py", "__pycache__"):
                continue
            entry_path = os.path.join(path, name)
            path_is_dir = True if os.path.isdir(entry_path) else False
            data = self.count_tests(entry_path)
            if data is None:
                continue
            if data[0] == 0:
                continue
            count_tests += data[0]
            if path_is_dir:
                directories.append(name)
            else:
                modules.append(name)
        return count_tests, directories, modules

    def tests_in_module(self, path):
        if not os.path.exists(path):
            return
        data = self.count_tests(path)
        if data is None:
            return
        count_tests, suite = data
        flatten_suite = self._flatten_suite(suite, [])
        classes = []
        for x in flatten_suite:
            if x.__class__ not in classes:
                classes.append(x.__class__)
        return count_tests, tuple([x.__name__ for x in classes])

    def tests_in_class(self, path, class_name):
        if not os.path.exists(path):
            return
        test_loader = unittest.TestLoader()
        data = self.count_tests(path, class_name=class_name, test_loader=test_loader)
        if data is None:
            return
        count_tests, suite = data
        flatten_suite = self._flatten_suite(suite, [])
        class_ = None
        for x in flatten_suite:
            if x.__class__.__name__ == class_name:
                class_ = x.__class__
                break
        if class_:
            test_cases = test_loader.getTestCaseNames(class_)
            return len(test_cases), test_cases

    def run(self, test_id, queue, path, class_name=None, method_name=None, failfast=False):
        data = self.count_tests(path, class_name=class_name, method_name=method_name)
        if data is None:
            return False
        count_tests, suite = data
        self.run_suite(test_id, suite, queue, failfast)

    def run_suite(self, test_id, suite, queue, failfast=False):
        self._reloader.restore_state()
        result = self._result_builder.build(test_id, queue)
        self._running_test[test_id] = dict()
        self._running_test[test_id]["result"] = result
        self._running_test[test_id]["suite"] = suite
        time_a = time.time()
        try:
            result.startTestRun()
            result.failfast = failfast
            suite.run(result)
        except Exception as e:
            return False
        finally:
            time_b = time.time()
            time_elapsed = time_b - time_a
            data = {"id": test_id, "event": "time_elapsed", "time": time_elapsed}
            queue.put(data)
            result.stopTestRun()

    def stop(self, test_id):
        try:
            self._running_test[test_id]["result"].stop()
        except Exception as e:
            pass

    def _setup(self):
        shared_folder = os.path.join(constant.PYRUSTIC_DATA_FOLDER, "jupitest")
        shared_json_path = os.path.join(shared_folder,
                                        "jupitest_shared_data.json")
        if not os.path.exists(shared_folder):
            os.makedirs(shared_folder)
        if not os.path.exists(shared_json_path):
            default_json = pkgutil.get_data("jupitest",
                                            "misc/default_shared_data.json")
            with open(shared_json_path, "wb") as file:
                file.write(default_json)
        self._jasonix = Jasonix(shared_json_path)

    def _flatten_suite(self, suite, result):
        if hasattr(suite, "__iter__"):
            for x in suite:
                self._flatten_suite(x, result)
        else:
            result.append(suite)
        return result

    def _convert_path_to_dotted_path(self, path, start=None):
        path = os.path.normpath(path)
        parts = path.split(os.sep)
        result = ""
        access = False if start else True
        for i, x in enumerate(parts):
            if not access:
                if x == start:
                    access = True
            if access:
                result += x
                if i < (len(parts) - 1):
                    result += "."
        return result



import tkinter as tk
from pyrustic.widget.tree import Tree as PyrusticTree


class Tree(PyrusticTree):
    def __init__(self, master, callback):
        super().__init__(master=master, spacing=20)
        self._callback = callback
        self._nodes = dict()

    def _on_build_node(self, header_frame, node):
        title = node["title"]
        node_id = node["node_id"]
        node_type = node["data"]["type"]
        container = node["container"]
        expander_stringvar = tk.StringVar()
        title_one_stringvar = tk.StringVar()
        title_two_stringvar = tk.StringVar()
        self._nodes[node_id] = {"expander_stringvar": expander_stringvar,
                                "title_one_stringvar": title_one_stringvar,
                                "title_two_stringvar": title_two_stringvar}
        # Titlebar
        titlebar = tk.Frame(header_frame)
        titlebar.grid(row=0, column=0, sticky="we")
        # Toolbar
        toolbar = tk.Frame(header_frame)
        toolbar.columnconfigure(0, weight=1)
        toolbar.grid(row=1, column=0, sticky="we")
        # expander
        if container:
            expander_stringvar.set("+")
            command = lambda self=self, node_id=node_id: self.collexp(node_id)
            expander = tk.Button(titlebar,
                                 name="treeExpanderButton",
                                 textvariable=expander_stringvar,
                                 command=command)
            expander.grid(row=0, column=0, sticky="w", padx=(0, 5))
        # titlelabel_one
        title_one_stringvar.set(node_type)
        titlelabel_one = tk.Label(titlebar,
                                  name="treeTitleLabelOne",
                                  textvariable=title_one_stringvar)
        titlelabel_one.grid(row=0, column=1, sticky="w", padx=(0, 5))
        # titlelabel_two
        title_two_stringvar.set(title)
        titlelabel_two = tk.Label(titlebar,
                                  name="treeTitleLabelTwo",
                                  textvariable=title_two_stringvar)
        titlelabel_two.grid(row=0, column=2, sticky="w")
        # binding command to titlebar, so callback will trigger toolbar
        command = (lambda event,
                          self=self,
                          toolbar=toolbar,
                          node=node: self._callback.on_title_clicked(node, toolbar))
        titlelabel_one.bind("<Button-1>", command)
        titlelabel_two.bind("<Button-1>", command)

    def _on_collapse(self, node):
        node_id = node["node_id"]
        self._nodes[node_id]["expander_stringvar"].set("+")
        self._callback.on_node_collapsed(node)

    def _on_expand(self, node):
        node_id = node["node_id"]
        self._nodes[node_id]["expander_stringvar"].set("-")
        self._callback.on_node_expanded(node)



from pyrustic.widget.scrollbox import Scrollbox
from pyrustic.viewable import Viewable
from jupitest.view.tree import Tree
import os
import os.path
import tkinter as tk


class MainView(Viewable):

    def __init__(self, app, threadom, host, toolbar_builder, root_path):
        self._app = app
        self._threadom = threadom
        self._host = host
        self._toolbar_builder = toolbar_builder
        self._root_path = root_path
        self._master = app.root
        self._focused_node = dict()
        self._test_running = False
        self._waiting_list = list()

    # ========================================
    #          INTERACTION WITH HOST
    # ========================================
    def fill_directory(self, node_id, data):
        self._tree.clear(node_id)
        if data is None:
            return
        count_tests, directories, modules = data
        self._tree.tag(node_id)["count_tests_stringvar"].set(count_tests)
        path = self._tree.node(node_id)["data"]["path"]
        for directory in directories:
            self._tree.insert(parent=node_id,
                              title= directory,
                              data={"path": os.path.join(path, directory),
                                    "type": "package",
                                    "count_tests_stringvar": tk.StringVar()})
        for module in modules:
            self._tree.insert(parent=node_id,
                              title= module,
                              data={"path": os.path.join(path, module),
                                    "type": "module",
                                    "count_tests_stringvar": tk.StringVar()})

    def fill_module(self, node_id, data):
        self._tree.clear(node_id)
        if data is None:
            return
        count_tests, classes = data
        if count_tests == 0:
            return
        self._tree.tag(node_id)["count_tests_stringvar"].set(count_tests)
        path = self._tree.node(node_id)["data"]["path"]
        for class_ in classes:
            self._tree.insert(parent=node_id,
                              title=class_,
                              data={"path": path,
                                    "type": "class",
                                    "name": class_,
                                    "count_tests_stringvar": tk.StringVar()})

    def fill_class(self, node_id, data):
        self._tree.clear(node_id)
        if data is None:
            return
        count_tests, methods = data
        self._tree.tag(node_id)["count_tests_stringvar"].set(count_tests)
        node = self._tree.node(node_id)
        module_path = node["data"]["path"]
        name = node["data"]["name"]
        for method in methods:
            self._tree.insert(parent=node_id,
                              title=method,
                              container=False,
                              data={"path": module_path,
                                    "type": "method",
                                    "class": name,
                                    "name": method,
                                    "count_tests_stringvar": tk.StringVar()})

    def notify_running_test_event(self, data):
        node_id = data["id"]
        event = data["event"]
        if node_id not in self._focused_node:
            return
        node = self._focused_node[node_id]
        toolbar = node["toolbar"]
        node["log"].append(data)
        if event == "start_test_run":
            self._test_running = True
            toolbar.set_running_state()
        elif event == "stop_test_run":
            self._test_running = False
            toolbar.set_stop_state(data["was_successful"], node["log"])
            node["log"] = []
            self._threadom.stop(node["qid"])
            if self._waiting_list:
                next_task = self._waiting_list[0]
                del self._waiting_list[0]
                self._run_test(*next_task)
        elif event == "time_elapsed":
            toolbar.notify_time_elapsed(data["time"])
        elif event == "start_test":
            toolbar.notify_start_test(data["test"])
        elif event == "stop_test":
            toolbar.notify_stop_test(data["test"])
        elif event == "add_error":
            toolbar.notify_add_error(data["test"], data["err"])
        elif event == "add_failure":
            toolbar.notify_add_failure(data["test"], data["err"])
        elif event == "add_success":
            toolbar.notify_add_success(data["test"])
        elif event == "add_skip":
            toolbar.notify_add_skip(data["test"], data["reason"])
        elif event == "add_expected_failure":
            toolbar.notify_add_expected_failure(data["test"], data["err"])
        elif event == "add_unexpected_success":
            toolbar.notify_add_unexpected_success(data["test"])
        elif event == "add_sub_test":
            toolbar.notify_add_sub_test(data["test"], data["sub_test"], data["outcome"])

    def notify_running_exception(self, exception):
        pass

    # ==================================
    #       INTERACTION WITH TREE
    # ==================================

    def on_node_collapsed(self, node):
        self._tree.clear(node["node_id"])
        node["box_frame"].config(height=1)

    def on_node_expanded(self, node):
        node_id = node["node_id"]
        node_type = node["data"]["type"]
        path = node["data"]["path"]

        if node_type in ("package", "root"):
            self._threadom.run(self._host.tests_in_directory, args=(path,),
                        consumer=lambda data, self=self, node_id=node_id:
                        self.fill_directory(node_id, data))
        elif node_type == "module":
            self._threadom.run(self._host.tests_in_module, args=(path,),
                           consumer=lambda data, self=self, node_id=node_id:
                           self.fill_module(node_id, data))
        elif node_type == "class":
            name = node["data"]["name"]
            self._threadom.run(self._host.tests_in_class, args=(path, name),
                           consumer=lambda data, self=self, node_id=node_id:
                           self.fill_class(node_id, data))

    def on_title_clicked(self, node, toolbar_parent):
        node_id = node["node_id"]
        toolbar = None
        operation = "create"
        if node_id in self._focused_node:
            toolbar = self._focused_node[node_id]["toolbar"]
            operation = "hide" if toolbar.is_visible else "show"
        if operation == "hide":
            toolbar.show_or_hide()
            self._focused_node[node_id]["parent"].config(height=1)
            return
        path = node["data"]["path"]
        class_name = method_name = None
        if node["data"]["type"] == "class":
            class_name = node["data"]["name"]
        elif node["data"]["type"] == "method":
            class_name = node["data"]["class"]
            method_name = node["data"]["name"]
        data = self._host.count_tests(path, class_name=class_name,
                                      method_name=method_name)
        if data is None:
            try:
                self._tree.collapse(node["parent"])
                self._tree.node(node["parent"])["box"].config(height=1)
            except Exception as e:
                pass
            return
        count_tests = data[0]
        if operation == "show":
            toolbar.count_tests = count_tests
            toolbar.show_or_hide()
            return
        toolbar = self._toolbar_builder.build(node_id, toolbar_parent, self)
        toolbar.count_tests = count_tests
        data = {"toolbar": toolbar, "parent": toolbar_parent,
                "running": False, "queue": None, "qid": None, "log": []}
        self._focused_node[node_id] = data


    # ========================================
    #          INTERACTION WITH TOOLBAR
    # ========================================
    def on_run_clicked(self, node_id, failfast):
        failfast = True if failfast else False
        toolbar = self._focused_node[node_id]["toolbar"]
        if self._test_running:
            toolbar.set_waiting_state()
            self._waiting_list.append((node_id, failfast))
        else:
            self._run_test(node_id, failfast)

    def on_stop_clicked(self, node_id):
        self._host.stop(node_id)

    def on_cancel_clicked(self, node_id):
        for x, data in enumerate(self._waiting_list):
            if node_id == data[0]:
                del self._waiting_list[x]


    # ===========================================
    #         VIEWABLE METHODS IMPLEMENTATION
    # ===========================================
    def _on_build(self):
        # == Widgets
        self._body = tk.Frame(self._master)
        self._scrollbox = Scrollbox(self._body)
        self._scrollbox.build_pack(expand=1,
                                   fill=tk.BOTH)
        self._tree = Tree(self._scrollbox.box, self)
        self._tree.build_pack(expand=1,
                              fill=tk.BOTH,
                              padx=(5, 50), pady=(5, 50))

    def _on_display(self):
        self._tree.insert(title="tests",
                          data={"path": self._root_path,
                                "type": "root",
                                "count_tests_stringvar": tk.StringVar()})

    def _on_destroy(self):
        pass

    # ==================================
    #          PRIVATE METHODS
    # ==================================
    def _delete_node(self, node_id):
        self._tree.delete(node_id)
        try:
            del self._focused_node[node_id]
        except KeyError:
            pass

    def _run_test(self, node_id, failfast):
        node = self._tree.node(node_id)
        path = node["data"]["path"]
        class_name = method_name = None
        if node["data"]["type"] == "class":
            class_name = node["data"]["name"]
        elif node["data"]["type"] == "method":
            class_name = node["data"]["class"]
            method_name = node["data"]["name"]
        data = self._host.count_tests(path, class_name=class_name, method_name=method_name)
        if data is None:
            self._tree.node(node["parent"])["box_frame"].config(height=1)
            self._tree.collapse(node["parent"])
            return
        count_tests, suite = data
        toolbar = self._focused_node[node_id]["toolbar"]
        toolbar.count_tests = count_tests
        queue = self._threadom.q()
        self._focused_node[node_id]["queue"] = queue
        self._threadom.run(self._host.run_suite, args=(node_id, suite, queue),
                             kwargs={"failfast": failfast})
        qid = self._threadom.consume(queue, consumer=self.notify_running_test_event,
                                exception_handler=self.notify_running_exception)
        self._focused_node[node_id]["qid"] = qid



from pyrustic.theme import Theme
from pyrustic import default_style
from pyrustic.theme.cyberpunk import Cyberpunk
from pyrustic.theme.cyberpunk import constant
from pyrustic.theme.cyberpunk.pyrustic_widget import tree
from pyrustic.theme.cyberpunk.native_widget import label
from pyrustic.theme.cyberpunk.native_widget import button


# ========================================
# RUNTEST THEME BASED ON CYBERPUNK THEME
# ========================================
def get_theme():
    theme = Cyberpunk()
    theme.add_theme(_get_tree_theme())
    theme.add_theme(_get_toolbar_theme())
    theme.add_theme(_get_log_window_theme())
    return theme


# ========================================
#                   TREE
# ========================================
# == Tree Theme
def _get_tree_theme():
    theme = tree.get_theme()
    theme.add_style(_get_tree_titlebar_style(), scope="*Tree*")
    theme.add_style(_get_tree_expander_button_style(), scope="*Tree*treeExpanderButton.")
    theme.add_style(_get_tree_title_one_label_style(), scope="*Tree*treeTitleLabelOne.")
    theme.add_style(_get_tree_title_two_label_style(), scope="*Tree*treeTitleLabelTwo.")
    return theme


# titlebar
def _get_tree_titlebar_style():
    style = default_style.Frame()
    style.background = constant.COLOR_BLACK
    return style


# expander button
def _get_tree_expander_button_style():
    style = default_style.Button()
    style.font = constant.FONT_FAV_BOLD
    style.background = constant.COLOR_BLACK
    style.foreground = "gray"
    style.activeForeground = constant.COLOR_BLACK
    style.highlightThickness = 0
    style.borderWidth = 0
    style.activeBackground = "#F0F8FF"
    style.padX = 3
    style.padY = 1
    return style


# title_one label
def _get_tree_title_one_label_style():
    style = default_style.Label()
    style.font = constant.FONT_FAV_BOLD
    style.background = constant.COLOR_BLACK
    style.foreground = "gray"
    return style


# title_two label
def _get_tree_title_two_label_style():
    style = default_style.Label()
    style.font = constant.FONT_FAV_BOLD
    style.background = constant.COLOR_BLACK
    style.foreground = "#CFCFCF"
    return style


# ========================================
#            TOOLBAR
# ========================================
# == toolbar theme
def _get_toolbar_theme():
    theme = Theme()
    theme.add_style(_get_toolbar_body_style(),
                    scope="*runtestToolbar*")
    theme.add_style(_get_toolbar_label_testing_passed_style(),
                    scope="*runtestToolbar*labelTestingPassed.")
    theme.add_style(_get_toolbar_label_testing_failed_style(),
                    scope="*runtestToolbar*labelTestingFailed.")
    theme.add_style(_get_toolbar_button_run(),
                    scope="*runtestToolbar*buttonRun.")
    theme.add_style(_get_toolbar_button_rerun(),
                    scope="*runtestToolbar*buttonRerun.")
    theme.add_style(_get_toolbar_button_stop(),
                    scope="*runtestToolbar*buttonStop.")
    theme.add_style(_get_toolbar_button_cancel(),
                    scope="*runtestToolbar*buttonCancel.")
    theme.add_style(_get_toolbar_button_clean(),
                    scope="*runtestToolbar*buttonClean.")
    theme.add_style(_get_toolbar_button_log(),
                    scope="*runtestToolbar*buttonLog.")
    return theme


# toolbar body
def _get_toolbar_body_style():
    style = default_style.Frame()
    style.background = "#002323"
    return style


# label testing passed
def _get_toolbar_label_testing_passed_style():
    style = label.get_style()
    style.foreground = "#40A640"
    style.background = "#002323"
    return style


# label testing failed
def _get_toolbar_label_testing_failed_style():
    style = label.get_style()
    style.foreground = "#F73030"
    style.background = "#002323"
    return style


# button run
def _get_toolbar_button_run():
    style = button.get_style()
    style.background = "#004600"
    style.foreground = constant.COLOR_ALMOST_WHITE
    style.activeBackground = "#006600"
    style.activeForeground = constant.COLOR_ALMOST_WHITE
    style.highlightBackground = constant.COLOR_ALMOST_WHITE
    style.highlightColor = constant.COLOR_ALMOST_WHITE
    return style


# button rerun
def _get_toolbar_button_rerun():
    style = _get_toolbar_button_run()
    return style


# button stop
def _get_toolbar_button_stop():
    style = _get_toolbar_button_run()
    style.background = "#CF0000"
    style.activeBackground = "#FF0000"
    return style


# button cancel
def _get_toolbar_button_cancel():
    style = _get_toolbar_button_run()
    style.background = "#BF2600"
    style.activeBackground = "#D73E18"
    return style


# button clean
def _get_toolbar_button_clean():
    style = _get_toolbar_button_run()
    style.background = constant.COLOR_BLACK
    style.activeBackground = "#FF0000"
    style.foreground = "gray"
    style.borderWidth = 0
    style.highlightThickness = 0
    return style


# button log
def _get_toolbar_button_log():
    style = _get_toolbar_button_run()
    style.background = "#003366"
    style.activeBackground = "#204B7E"
    return style


# ========================================
#               LOG WINDOW
# ========================================
# == log window theme
def _get_log_window_theme():
    theme = Theme()
    theme.add_style(_get_log_window_text_style(), scope="*Text.")
    return theme


# log window
def _get_log_window_text_style():
    style = default_style.Text()
    style.font = constant.FONT_DEFAULT_FAMILY, 15, "normal"
    style.background = "#033669"
    style.foreground = "#7EB1B1"
    style.highlightThickness = 0
    return style



import unittest


class Result(unittest.TestResult):

    def __init__(self, test_id, queue):
        super().__init__()
        self._test_id = test_id
        self._queue = queue

    def startTestRun(self):
        super().startTestRun()
        data = {"id": self._test_id, "event": "start_test_run"}
        self._queue.put(data)

    def stopTestRun(self):
        super().stopTestRun()
        data = {"id": self._test_id, "event": "stop_test_run",
                "was_successful": self.wasSuccessful()}
        self._queue.put(data)

    def startTest(self, test):
        super().startTest(test)
        data = {"id": self._test_id, "event": "start_test",
                "test": test}
        self._queue.put(data)

    def stopTest(self, test):
        super().stopTest(test)
        data = {"id": self._test_id, "event": "stop_test",
                "test": test}
        self._queue.put(data)

    def addError(self, test, err):
        """'err' is a tuple of values as returned by sys.exc_info()"""
        super().addError(test, err)
        data = {"id": self._test_id, "event": "add_error",
                "test": test, "err": err}
        self._queue.put(data)

    def addFailure(self, test, err):
        super().addFailure(test, err)
        data = {"id": self._test_id, "event": "add_failure",
                "test": test, "err": err}
        self._queue.put(data)

    def addSuccess(self, test):
        super().addSuccess(test)
        data = {"id": self._test_id, "event": "add_success",
                "test": test}
        self._queue.put(data)

    def addSkip(self, test, reason):
        super().addSkip(test, reason)
        data = {"id": self._test_id, "event": "add_skip",
                "test": test, "reason": reason}
        self._queue.put(data)

    def addExpectedFailure(self, test, err):
        super().addExpectedFailure(test, err)
        data = {"id": self._test_id, "event": "add_expected_failure",
                "test": test, "err": err}
        self._queue.put(data)

    def addUnexpectedSuccess(self, test):
        super().addUnexpectedSuccess(test)
        data = {"id": self._test_id, "event": "add_unexpected_success",
                "test": test}
        self._queue.put(data)

    def addSubTest(self, test, subtest, outcome):
        try:  # new in Python version 3.4
            super().addSubTest(test, subtest, outcome)
        except Exception as e:
            return
        data = {"id": self._test_id, "event": "add_sub_test",
                "test": test, "sub_test": subtest, "outcome": outcome}
        self._queue.put(data)
